public with sharing class OrdersController {
	
	// Debugging Constant
	private static final string TAG = '***OrdersController.cls - ';
	
	// Order for the current context
	public Order__c CurrentOrder{get; set;}
	
	// Order Lines for current Order
	public List<Order_Line__c> OrderLines{get; set;}
		
	// public property for inserting new Order Line record
	public Order_Line__c NewOrderLine {get; set;}
	
	// public property for the Order Line being edited
	public Id editOrderLineId {get; set;}	
	
	// public property for the Order Line being removed
	public Id removeOrderLineId {get; set;}
	
	// flag whether to render Order Header only
	public boolean renderOrderDetails {get; set;}
	
	// flag whether any changes have been made to order
	public boolean isOrderDirty {get; set;}
			
	// constructor
	public OrdersController()
	{
		// initialize CurrentOrder property
		CurrentOrder = new Order__c();
				
		// initialize property in case we want to add a new order line
		NewOrderLine = new Order_Line__c();
		
		// initialize editOrderLineId to null
		editOrderLineId = null;
				
		// get the Order Lines
		prepareFullOrder();
		
		// no changes to order at this point
		isOrderDirty = false;
		
	}
	
	// get a List of Order Lines for the current order
	public void prepareFullOrder()
	{
		// get the Current Order using id page parameter
		String id = ApexPages.currentPage().getParameters().get('id');
		if (id <> null) {
			// attempt retrieve existing order
			try {
				// render all components by default
				renderOrderDetails = true;
				
				
				// attempt to retrieve order using the passed in id
				CurrentOrder = [select Id, Name, Delivered__c, Planned_Delivery_Date__c,
											Order_Lines_Total__c, Discount__c,
											Comments__c, Grand_Total__c,
											Customer_Name__c,
											Customer_Email__c,
											Channel__c,
											Customer_Address__c 
								from Order__c where id = :id];
		
				// Get the Order Lines for the Order
				OrderLines = new List<Order_Line__c>();
				OrderLines = [Select Name, Order_Line_Item__c, 
								Line_Item_Price__c, Line_Item_Total__c, 
								Order_Line_Item__r.Item_Name__c,
								Order__c, Quantity__c
								FROM Order_Line__c WHERE Order__c = :this.CurrentOrder.Id
								ORDER BY Name ASC];
			} catch (QueryException ex) {
				ApexPages.Message myMsg = new ApexPages.Message(ApexPages.Severity.ERROR,'Order ID does not exist or is not unique!');
				ApexPages.addMessage(myMsg);
				
			} catch (Exception ex) {
				ApexPages.addMessages(ex);
			}
			
		} else {
			// we are adding a new Order
			renderOrderDetails = false;
		}		
	}
	
	// save the Order Header information
	public PageReference saveOrderHeader() {
		PageReference pageRef = null;
		string methodName = 'saveOrderHeader - ';
		System.debug(TAG + methodName + 'Inserting/Updating Order ' + '...');
		try {
			if (CurrentOrder.Id == null) 
			{
				// add record and reset flags
				insert CurrentOrder;
				System.debug('Inserted Order# ' + CurrentOrder.Id + '...');
				renderOrderDetails = true;
				// set up page reference to display newly added order
				pageRef = ApexPages.currentPage();
				pageRef.getParameters().put('id',CurrentOrder.Id);
				pageRef.setRedirect(true);
			} else {
				// update the current order
				update CurrentOrder;
			}
			// reset flag and refresh order details in controller
			isOrderDirty = false;
			prepareFullOrder();
			System.debug(TAG + methodName + 'Successfully Inserted/Updated Order ' + CurrentOrder.Name + '...');
			successMessage('Order Successfully Saved!');
		} catch (DmlException ex) {
			ApexPages.addMessages(ex);
		}
		return pageRef;
	}
	
	// Cancel the Order
	public PageReference cancelOrder() {
		string methodName = 'cancelOrder - ';
		System.debug(TAG + methodName + 'Cancelling Order #' + CurrentOrder.Name + '...');
		try {
			// Delete Order
			delete CurrentOrder; 
		} catch (DmlException ex) {
			ApexPages.addMessages(ex);
			return null;
		}
		// re-direct back to Orders tab
		return closeOrder();
	}
	
	// Insert a new Order Line
	public PageReference insertOrderLine() {
		string methodName = 'insertOrderLine - ';
		System.debug(TAG + methodName + 'Inserting Order Line for Item #' + NewOrderLine.Order_Line_Item__c + '...');
		try {
			NewOrderLine.Order__c = this.CurrentOrder.Id;
			insert NewOrderLine;
			// re-display Order Line Items
			prepareFullOrder();
			// reset public property for new insert
			NewOrderLine = new Order_Line__c();
			successMessage('Order Line Successfully Added!');
			System.debug(TAG + methodName + 'Successfully Inserted New Order Line...');
		} catch (DmlException ex) {
			ApexPages.addMessages(ex);
		}
		return null;	// no page re-direct
	}
	
	// delete an Order Line
	public PageReference removeOrderLine() {
		string methodName = 'removeOrderLine - ';
		System.debug(TAG + methodName + 'Removing Order Line #' + removeOrderLineId + '...');
		try {
			Order_Line__c lineToDelete = [SELECT Id FROM Order_Line__c 
											WHERE Id = :removeOrderLineId
											LIMIT 1];
			delete lineToDelete;
			prepareFullOrder();
			successMessage('Order Line Successfully Removed!');
			removeOrderLineId = null;
		} catch (DmlException ex) {
			ApexPages.addMessages(ex);
		}
		return null;		
	}
	
	// Save the changes to an Order Line
	public PageReference saveOrderLine()
	{
		string methodName = 'saveOrderLines - ';
		System.debug(TAG + methodName + 'Saving Order Lines...');
		
		try {
			update OrderLines;
			successMessage('Order Line Successfully Saved!');
		} catch (DmlException ex) {
			System.debug(TAG + methodName  + ex);
			ApexPages.addMessages(ex);
			return null;
		}
		
		return cancelEditOrderLine();
	}
	
	// cancel editing of an order line
	public PageReference cancelEditOrderLine()
	{
		string methodName = 'cancelEditOrderLine - ';
		System.debug(TAG + methodName + 'Cancelling Edit of Order Line...');
		editOrderLineId = null;
		prepareFullOrder();
		return null;
	}
	
	// return page reference to Orders tab
	public PageReference closeOrder()
	{
		// Note this may differ for your org
		return new PageReference('/apex/OrderSearchPage');
	}
	
	// display a page message as required
	private void successMessage(string messageToDisplay) {
		ApexPages.Message myMsg = new ApexPages.Message(ApexPages.Severity.INFO,messageToDisplay);
		ApexPages.addMessage(myMsg);		
	}

}